"""Tests the full end-to-end extraction process.

Set the bash env variable `DO_USE_LOCAL_SCRIPTS=1` to use the local py files, rather than the installed
scripts.
"""

import json
import tempfile
from datetime import datetime
from io import StringIO
from pathlib import Path

import polars as pl
from meds import __version__ as MEDS_VERSION
from meds import code_metadata_filepath, dataset_metadata_filepath, subject_splits_filepath

from tests.MEDS_Extract import (
    CONVERT_TO_SHARDED_EVENTS_SCRIPT,
    EXTRACT_CODE_METADATA_SCRIPT,
    FINALIZE_DATA_SCRIPT,
    FINALIZE_METADATA_SCRIPT,
    MERGE_TO_MEDS_COHORT_SCRIPT,
    SHARD_EVENTS_SCRIPT,
    SPLIT_AND_SHARD_SCRIPT,
)
from tests.utils import assert_df_equal, run_command

# Test data (inputs)

SUBJECTS_CSV = """
MRN,dob,eye_color,height
1195293,06/20/1978,BLUE,164.6868838269085
239684,12/28/1980,BROWN,175.271115221764
1500733,07/20/1986,BROWN,158.60131573580904
814703,03/28/1976,HAZEL,156.48559093209357
754281,12/19/1988,BROWN,166.22261567137025
68729,03/09/1978,HAZEL,160.3953106166676
"""

ADMIT_VITALS_CSV = """
subject_id,admit_date,disch_date,department,vitals_date,HR,temp
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 18:57:18",112.6,95.5
754281,"01/03/2010, 06:27:59","01/03/2010, 08:22:13",PULMONARY,"01/03/2010, 06:27:59",142.0,99.8
814703,"02/05/2010, 05:55:39","02/05/2010, 07:02:30",ORTHOPEDIC,"02/05/2010, 05:55:39",170.2,100.1
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 18:25:35",113.4,95.8
68729,"05/26/2010, 02:30:56","05/26/2010, 04:51:52",PULMONARY,"05/26/2010, 02:30:56",86.0,97.8
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 20:12:31",112.5,99.8
1500733,"06/03/2010, 14:54:38","06/03/2010, 16:44:26",ORTHOPEDIC,"06/03/2010, 16:20:49",90.1,100.1
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 17:48:48",105.1,96.2
239684,"05/11/2010, 17:41:51","05/11/2010, 19:27:19",CARDIAC,"05/11/2010, 17:41:51",102.6,96.0
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 19:25:32",114.1,100.0
1500733,"06/03/2010, 14:54:38","06/03/2010, 16:44:26",ORTHOPEDIC,"06/03/2010, 14:54:38",91.4,100.0
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 20:41:33",107.5,100.4
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 20:24:44",107.7,100.0
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 19:45:19",119.8,99.9
1195293,"06/20/2010, 19:23:52","06/20/2010, 20:50:04",CARDIAC,"06/20/2010, 19:23:52",109.0,100.0
1500733,"06/03/2010, 14:54:38","06/03/2010, 16:44:26",ORTHOPEDIC,"06/03/2010, 15:39:49",84.4,100.3
"""

INPUT_METADATA_FILE = """
lab_code,title,loinc
HR,Heart Rate,8867-4
temp,Body Temperature,8310-5
"""

DEMO_METADATA_FILE = """
eye_color,description
BROWN,"Brown Eyes. The most common eye color."
BLUE,"Blue Eyes. Less common than brown."
HAZEL,"Hazel eyes. These are uncommon"
GREEN,"Green eyes. These are rare."
"""

EVENT_CFGS_YAML = """
subjects:
  subject_id_col: MRN
  eye_color:
    code:
      - EYE_COLOR
      - col(eye_color)
    time: null
    _metadata:
      demo_metadata:
        description: description
  height:
    code: HEIGHT
    time: null
    numeric_value: height
  dob:
    code: DOB
    time: col(dob)
    time_format: "%m/%d/%Y"
admit_vitals:
  admissions:
    code:
      - ADMISSION
      - col(department)
    time: col(admit_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
  discharge:
    code: DISCHARGE
    time: col(disch_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
  HR:
    code: HR
    time: col(vitals_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
    numeric_value: HR
    _metadata:
      input_metadata:
        description: {"title": {"lab_code": "HR"}}
        parent_codes: {"LOINC/{loinc}": {"lab_code": "HR"}}
  temp:
    code: TEMP
    time: col(vitals_date)
    time_format: "%m/%d/%Y, %H:%M:%S"
    numeric_value: temp
    _metadata:
      input_metadata:
        description: {"title": {"lab_code": "temp"}}
        parent_codes: {"LOINC/{loinc}": {"lab_code": "temp"}}
"""

# Test data (expected outputs) -- ALL OF THIS MAY CHANGE IF THE SEED OR DATA CHANGES
EXPECTED_SPLITS = {
    "train/0": [239684, 1195293],
    "train/1": [68729, 814703],
    "tuning/0": [754281],
    "held_out/0": [1500733],
}

SUBJECT_SPLITS_DF = pl.DataFrame(
    {
        "subject_id": [239684, 1195293, 68729, 814703, 754281, 1500733],
        "split": ["train", "train", "train", "train", "tuning", "held_out"],
    }
)


def get_expected_output(df: str) -> pl.DataFrame:
    return (
        pl.read_csv(source=StringIO(df))
        .select(
            "subject_id",
            pl.col("time").str.strptime(pl.Datetime, "%m/%d/%Y, %H:%M:%S").alias("time"),
            pl.col("code"),
            "numeric_value",
        )
        .sort(by=["subject_id", "time"])
    )


MEDS_OUTPUT_TRAIN_0_SUBJECTS = """
subject_id,time,code,numeric_value
239684,,EYE_COLOR//BROWN,
239684,,HEIGHT,175.271115221764
239684,"12/28/1980, 00:00:00",DOB,
1195293,,EYE_COLOR//BLUE,
1195293,,HEIGHT,164.6868838269085
1195293,"06/20/1978, 00:00:00",DOB,
"""

MEDS_OUTPUT_TRAIN_0_ADMIT_VITALS = """
subject_id,time,code,numeric_value
239684,"05/11/2010, 17:41:51",ADMISSION//CARDIAC,
239684,"05/11/2010, 17:41:51",HR,102.6
239684,"05/11/2010, 17:41:51",TEMP,96.0
239684,"05/11/2010, 17:48:48",HR,105.1
239684,"05/11/2010, 17:48:48",TEMP,96.2
239684,"05/11/2010, 18:25:35",HR,113.4
239684,"05/11/2010, 18:25:35",TEMP,95.8
239684,"05/11/2010, 18:57:18",HR,112.6
239684,"05/11/2010, 18:57:18",TEMP,95.5
239684,"05/11/2010, 19:27:19",DISCHARGE,
1195293,"06/20/2010, 19:23:52",ADMISSION//CARDIAC,
1195293,"06/20/2010, 19:23:52",HR,109.0
1195293,"06/20/2010, 19:23:52",TEMP,100.0
1195293,"06/20/2010, 19:25:32",HR,114.1
1195293,"06/20/2010, 19:25:32",TEMP,100.0
1195293,"06/20/2010, 19:45:19",HR,119.8
1195293,"06/20/2010, 19:45:19",TEMP,99.9
1195293,"06/20/2010, 20:12:31",HR,112.5
1195293,"06/20/2010, 20:12:31",TEMP,99.8
1195293,"06/20/2010, 20:24:44",HR,107.7
1195293,"06/20/2010, 20:24:44",TEMP,100.0
1195293,"06/20/2010, 20:41:33",HR,107.5
1195293,"06/20/2010, 20:41:33",TEMP,100.4
1195293,"06/20/2010, 20:50:04",DISCHARGE,
"""

MEDS_OUTPUT_TRAIN_1_SUBJECTS = """
subject_id,time,code,numeric_value
68729,,EYE_COLOR//HAZEL,
68729,,HEIGHT,160.3953106166676
68729,"03/09/1978, 00:00:00",DOB,
814703,,EYE_COLOR//HAZEL,
814703,,HEIGHT,156.48559093209357
814703,"03/28/1976, 00:00:00",DOB,
"""

MEDS_OUTPUT_TRAIN_1_ADMIT_VITALS = """
subject_id,time,code,numeric_value
68729,"05/26/2010, 02:30:56",ADMISSION//PULMONARY,
68729,"05/26/2010, 02:30:56",HR,86.0
68729,"05/26/2010, 02:30:56",TEMP,97.8
68729,"05/26/2010, 04:51:52",DISCHARGE,
814703,"02/05/2010, 05:55:39",ADMISSION//ORTHOPEDIC,
814703,"02/05/2010, 05:55:39",HR,170.2
814703,"02/05/2010, 05:55:39",TEMP,100.1
814703,"02/05/2010, 07:02:30",DISCHARGE,
"""

MEDS_OUTPUT_TUNING_0_SUBJECTS = """
subject_id,time,code,numeric_value
754281,,EYE_COLOR//BROWN,
754281,,HEIGHT,166.22261567137025
754281,"12/19/1988, 00:00:00",DOB,
"""

MEDS_OUTPUT_TUNING_0_ADMIT_VITALS = """
subject_id,time,code,numeric_value
754281,"01/03/2010, 06:27:59",ADMISSION//PULMONARY,
754281,"01/03/2010, 06:27:59",HR,142.0
754281,"01/03/2010, 06:27:59",TEMP,99.8
754281,"01/03/2010, 08:22:13",DISCHARGE,
"""

MEDS_OUTPUT_HELD_OUT_0_SUBJECTS = """
subject_id,time,code,numeric_value
1500733,,EYE_COLOR//BROWN,
1500733,,HEIGHT,158.60131573580904
1500733,"07/20/1986, 00:00:00",DOB,
"""

MEDS_OUTPUT_HELD_OUT_0_ADMIT_VITALS = """
subject_id,time,code,numeric_value
1500733,"06/03/2010, 14:54:38",ADMISSION//ORTHOPEDIC,
1500733,"06/03/2010, 14:54:38",HR,91.4
1500733,"06/03/2010, 14:54:38",TEMP,100.0
1500733,"06/03/2010, 15:39:49",HR,84.4
1500733,"06/03/2010, 15:39:49",TEMP,100.3
1500733,"06/03/2010, 16:20:49",HR,90.1
1500733,"06/03/2010, 16:20:49",TEMP,100.1
1500733,"06/03/2010, 16:44:26",DISCHARGE,
"""

MEDS_OUTPUT_CODE_METADATA_FILE = """
code,description,parent_codes
EYE_COLOR//BLUE,"Blue Eyes. Less common than brown.",
EYE_COLOR//BROWN,"Brown Eyes. The most common eye color.",
EYE_COLOR//HAZEL,"Hazel eyes. These are uncommon",
HR,"Heart Rate",LOINC/8867-4
TEMP,"Body Temperature",LOINC/8310-5
"""

MEDS_OUTPUT_DATASET_METADATA_JSON = {
    "dataset_name": "TEST",
    "dataset_version": "1.0",
    "etl_name": "MEDS_transforms",
    # "etl_version": None,  # We don't test this as it changes with the commits.
    "meds_version": MEDS_VERSION,
}

SUB_SHARDED_OUTPUTS = {
    "train/0": {
        "subjects": MEDS_OUTPUT_TRAIN_0_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_TRAIN_0_ADMIT_VITALS,
    },
    "train/1": {
        "subjects": MEDS_OUTPUT_TRAIN_1_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_TRAIN_1_ADMIT_VITALS,
    },
    "tuning/0": {
        "subjects": MEDS_OUTPUT_TUNING_0_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_TUNING_0_ADMIT_VITALS,
    },
    "held_out/0": {
        "subjects": MEDS_OUTPUT_HELD_OUT_0_SUBJECTS,
        "admit_vitals": MEDS_OUTPUT_HELD_OUT_0_ADMIT_VITALS,
    },
}


MEDS_OUTPUTS = {
    "train/0": [MEDS_OUTPUT_TRAIN_0_SUBJECTS, MEDS_OUTPUT_TRAIN_0_ADMIT_VITALS],
    "train/1": [MEDS_OUTPUT_TRAIN_1_SUBJECTS, MEDS_OUTPUT_TRAIN_1_ADMIT_VITALS],
    "tuning/0": [MEDS_OUTPUT_TUNING_0_SUBJECTS, MEDS_OUTPUT_TUNING_0_ADMIT_VITALS],
    "held_out/0": [MEDS_OUTPUT_HELD_OUT_0_SUBJECTS, MEDS_OUTPUT_HELD_OUT_0_ADMIT_VITALS],
}


def test_extraction():
    with tempfile.TemporaryDirectory() as d:
        raw_cohort_dir = Path(d) / "raw_cohort"
        MEDS_cohort_dir = Path(d) / "MEDS_cohort"

        # Create the directories
        raw_cohort_dir.mkdir()
        MEDS_cohort_dir.mkdir()

        subjects_csv = raw_cohort_dir / "subjects.csv"
        admit_vitals_csv = raw_cohort_dir / "admit_vitals.csv"
        event_cfgs_yaml = raw_cohort_dir / "event_cfgs.yaml"

        demo_metadata_csv = raw_cohort_dir / "demo_metadata.csv"
        input_metadata_csv = raw_cohort_dir / "input_metadata.csv"

        # Write the CSV files
        subjects_csv.write_text(SUBJECTS_CSV.strip())
        admit_vitals_csv.write_text(ADMIT_VITALS_CSV.strip())
        demo_metadata_csv.write_text(DEMO_METADATA_FILE.strip())
        input_metadata_csv.write_text(INPUT_METADATA_FILE.strip())

        # Mix things up -- have one CSV be also in parquet format.
        admit_vitals_parquet = raw_cohort_dir / "admit_vitals.parquet"
        df = pl.read_csv(admit_vitals_csv)

        df.write_parquet(admit_vitals_parquet, use_pyarrow=True)

        # Write the event config YAML
        event_cfgs_yaml.write_text(EVENT_CFGS_YAML)

        # Run the extraction script
        #   1. Sub-shard the data (this will be a null operation in this case, but it is worth doing just in
        #      case.
        #   2. Collect the subject splits.
        #   3. Extract the events and sub-shard by subject.
        #   4. Merge to the final output.

        extraction_config_kwargs = {
            "input_dir": str(raw_cohort_dir.resolve()),
            "cohort_dir": str(MEDS_cohort_dir.resolve()),
            "event_conversion_config_fp": str(event_cfgs_yaml.resolve()),
            "stage_configs.split_and_shard_subjects.split_fracs.train": 4 / 6,
            "stage_configs.split_and_shard_subjects.split_fracs.tuning": 1 / 6,
            "stage_configs.split_and_shard_subjects.split_fracs.held_out": 1 / 6,
            "stage_configs.shard_events.row_chunksize": 10,
            "stage_configs.split_and_shard_subjects.n_subjects_per_shard": 2,
            "hydra.verbose": True,
            "etl_metadata.dataset_name": "TEST",
            "etl_metadata.dataset_version": "1.0",
        }

        all_stderrs = []
        all_stdouts = []

        # Stage 1: Sub-shard the data
        stderr, stdout = run_command(SHARD_EVENTS_SCRIPT, extraction_config_kwargs, "shard_events")

        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        subsharded_dir = MEDS_cohort_dir / "shard_events"

        try:
            out_files = list(subsharded_dir.glob("**/*.parquet"))
            assert len(out_files) == 3, f"Expected 3 output files, got {len(out_files)}."

            # Checking specific out files:
            #   1. subjects.parquet
            subjects_out = subsharded_dir / "subjects" / "[0-6).parquet"
            assert subjects_out.is_file(), f"Expected {subjects_out} to exist. Files include {out_files}."

            assert_df_equal(
                pl.read_csv(subjects_csv),
                pl.read_parquet(subjects_out, glob=False),
                "Subjects should be equal after sub-sharding",
                check_column_order=False,
                check_row_order=False,
            )
        except AssertionError as e:
            full_stderr = "\n".join(all_stderrs)
            print("Sub-sharding failed")
            print(f"stderr:\n{full_stderr}")
            raise e

        #   2. admit_vitals.parquet
        df_chunks = []
        for chunk in ["[0-10)", "[10-16)"]:
            admit_vitals_chunk_fp = subsharded_dir / "admit_vitals" / f"{chunk}.parquet"
            assert admit_vitals_chunk_fp.is_file(), f"Expected {admit_vitals_chunk_fp} to exist."

            df_chunks.append(pl.read_parquet(admit_vitals_chunk_fp, glob=False))

        assert_df_equal(
            pl.read_csv(admit_vitals_csv),
            pl.concat(df_chunks),
            "Admit vitals should be equal after sub-sharding",
            check_column_order=False,
            check_row_order=False,
        )

        # Stage 2: Collect the subject splits
        stderr, stdout = run_command(
            SPLIT_AND_SHARD_SCRIPT,
            extraction_config_kwargs,
            "split_and_shard_subjects",
        )

        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        try:
            shards_fp = MEDS_cohort_dir / "metadata" / ".shards.json"
            assert shards_fp.is_file(), f"Expected splits @ {str(shards_fp.resolve())} to exist."

            splits = json.loads(shards_fp.read_text())
            expected_keys = ["train/0", "train/1", "tuning/0", "held_out/0"]

            expected_keys_str = ", ".join(f"'{k}'" for k in expected_keys)
            got_keys_str = ", ".join(f"'{k}'" for k in splits.keys())

            assert set(splits.keys()) == set(expected_keys), (
                f"Expected splits to have keys {expected_keys_str}.\n" f"Got keys: {got_keys_str}"
            )

            assert splits == EXPECTED_SPLITS, (
                f"Expected splits to be {EXPECTED_SPLITS}, got {splits}. NOTE THIS MAY CHANGE IF THE SEED OR "
                "DATA CHANGES -- FAILURE HERE MAY BE JUST DUE TO A NON-DETERMINISTIC SPLIT AND THE TEST "
                "NEEDING TO BE UPDATED."
            )
        except AssertionError as e:
            print("Failed to split subjects")
            print(f"stderr:\n{stderr}")
            print(f"stdout:\n{stdout}")
            raise e

        # Stage 3: Extract the events and sub-shard by subject
        stderr, stdout = run_command(
            CONVERT_TO_SHARDED_EVENTS_SCRIPT,
            extraction_config_kwargs,
            "convert_events",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        subject_subsharded_folder = MEDS_cohort_dir / "convert_to_sharded_events"
        assert subject_subsharded_folder.is_dir(), f"Expected {subject_subsharded_folder} to be a directory."

        for split, expected_outputs in SUB_SHARDED_OUTPUTS.items():
            for prefix, expected_df_L in expected_outputs.items():
                if not isinstance(expected_df_L, list):
                    expected_df_L = [expected_df_L]

                expected_df = pl.concat([get_expected_output(df) for df in expected_df_L])

                fps = list((subject_subsharded_folder / split / prefix).glob("*.parquet"))
                assert len(fps) > 0

                # We add a "unique" here as there may be some duplicates across the row-group sub-shards.
                got_df = pl.concat([pl.read_parquet(fp, glob=False) for fp in fps]).unique()
                try:
                    assert_df_equal(
                        expected_df,
                        got_df,
                        f"Expected output for split {split}/{prefix} to be equal to the expected output.",
                        check_column_order=False,
                        check_row_order=False,
                    )
                except AssertionError as e:
                    print(f"Failed on split {split}/{prefix}")
                    print(f"stderr:\n{stderr}")
                    print(f"stdout:\n{stdout}")
                    raise e

        # Stage 4: Merge to the final output
        stderr, stdout = run_command(
            MERGE_TO_MEDS_COHORT_SCRIPT,
            extraction_config_kwargs,
            "merge_to_MEDS_cohort",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        # Check the final output
        output_folder = MEDS_cohort_dir / "merge_to_MEDS_cohort"
        try:
            for split, expected_df_L in MEDS_OUTPUTS.items():
                if not isinstance(expected_df_L, list):
                    expected_df_L = [expected_df_L]

                expected_df = pl.concat([get_expected_output(df) for df in expected_df_L])

                fp = output_folder / f"{split}.parquet"
                assert fp.is_file(), f"Expected {fp} to exist.\nstderr:\n{stderr}\nstdout:\n{stdout}"

                got_df = pl.read_parquet(fp, glob=False)
                assert_df_equal(
                    expected_df,
                    got_df,
                    f"Expected output for split {split} to be equal to the expected output.",
                    check_column_order=False,
                    check_row_order=False,
                )

                assert got_df["subject_id"].is_sorted(), f"Subject IDs should be sorted for split {split}."
                for subj in splits[split]:
                    got_df_subj = got_df.filter(pl.col("subject_id") == subj)
                    assert got_df_subj[
                        "time"
                    ].is_sorted(), f"Times should be sorted for subject {subj} in split {split}."

        except AssertionError as e:
            print(f"Failed on split {split}")
            print(f"stderr:\n{full_stderr}")
            print(f"stdout:\n{full_stdout}")
            raise e

        # Stage 6: Extract code metadata
        stderr, stdout = run_command(
            EXTRACT_CODE_METADATA_SCRIPT,
            extraction_config_kwargs,
            "extract_code_metadata",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        output_file = MEDS_cohort_dir / "extract_code_metadata" / "codes.parquet"
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_df = pl.read_parquet(output_file, glob=False)

        want_df = pl.read_csv(source=StringIO(MEDS_OUTPUT_CODE_METADATA_FILE)).with_columns(
            pl.col("parent_codes").cast(pl.List(pl.Utf8)),
        )

        assert_df_equal(
            want=want_df,
            got=got_df,
            msg="Code metadata with descriptions differs!",
            check_column_order=False,
            check_row_order=False,
        )

        # Stage 7: Finalize the MEDS data
        stderr, stdout = run_command(
            FINALIZE_DATA_SCRIPT,
            extraction_config_kwargs,
            "finalize_MEDS_data",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        # Check the final output
        output_folder = MEDS_cohort_dir / "data"
        try:
            for split, expected_df_L in MEDS_OUTPUTS.items():
                if not isinstance(expected_df_L, list):
                    expected_df_L = [expected_df_L]

                expected_df = pl.concat([get_expected_output(df) for df in expected_df_L]).with_columns(
                    pl.col("numeric_value").cast(pl.Float32)
                )

                fp = output_folder / f"{split}.parquet"
                assert fp.is_file(), f"Expected {fp} to exist."

                got_df = pl.read_parquet(fp, glob=False)
                assert_df_equal(
                    expected_df,
                    got_df,
                    f"Expected output for split {split} to be equal to the expected output.",
                    check_column_order=False,
                    check_row_order=False,
                )

                assert got_df["subject_id"].is_sorted(), f"Subject IDs should be sorted for split {split}."
                for subj in splits[split]:
                    got_df_subj = got_df.filter(pl.col("subject_id") == subj)
                    assert got_df_subj[
                        "time"
                    ].is_sorted(), f"Times should be sorted for subject {subj} in split {split}."

        except AssertionError as e:
            print(f"Failed on split {split}")
            print(f"stderr:\n{full_stderr}")
            print(f"stdout:\n{full_stdout}")
            raise e

        # Stage 8: Finalize the metadata
        stderr, stdout = run_command(
            FINALIZE_METADATA_SCRIPT,
            extraction_config_kwargs,
            "finalize_metadata",
        )
        all_stderrs.append(stderr)
        all_stdouts.append(stdout)

        full_stderr = "\n".join(all_stderrs)
        full_stdout = "\n".join(all_stdouts)

        # Check code metadata
        output_file = MEDS_cohort_dir / code_metadata_filepath
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_df = pl.read_parquet(output_file, glob=False, use_pyarrow=True)

        want_df = pl.read_csv(source=StringIO(MEDS_OUTPUT_CODE_METADATA_FILE)).with_columns(
            pl.col("code"),
            pl.col("parent_codes").cast(pl.List(pl.Utf8)),
        )

        # We collapse the list type as it throws an error in the assert_df_equal otherwise
        got_df = got_df.with_columns(pl.col("parent_codes").list.join("||"))
        want_df = want_df.with_columns(pl.col("parent_codes").list.join("||"))

        assert_df_equal(
            want=want_df,
            got=got_df,
            msg=f"Finalized code metadata differs:\nstderr:\n{stderr}\nstdout:\n{stdout}",
            check_column_order=False,
            check_row_order=False,
        )

        # Check dataset metadata
        output_file = MEDS_cohort_dir / dataset_metadata_filepath
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_json = json.loads(output_file.read_text())
        assert "etl_version" in got_json, "Expected 'etl_version' to be in the dataset metadata."
        got_json.pop("etl_version")  # We don't test this as it changes with the commits.

        assert "created_at" in got_json, "Expected 'created_at' to be in the dataset metadata."
        created_at_obs = got_json.pop("created_at")
        as_dt = datetime.fromisoformat(created_at_obs)
        assert as_dt < datetime.now(), f"Expected 'created_at' to be before now, got {created_at_obs}."
        created_ago = datetime.now() - as_dt
        assert created_ago.total_seconds() < 5 * 60, "Expected 'created_at' to be within 5 minutes of now."

        assert got_json == MEDS_OUTPUT_DATASET_METADATA_JSON, f"Dataset metadata differs: {got_json}"

        output_file = MEDS_cohort_dir / subject_splits_filepath
        assert output_file.is_file(), f"Expected {output_file} to exist: stderr:\n{stderr}\nstdout:\n{stdout}"

        got_df = pl.read_parquet(output_file, glob=False, use_pyarrow=True)
        assert_df_equal(
            SUBJECT_SPLITS_DF,
            got_df,
            "Subject splits should be equal to the expected splits.",
            check_column_order=False,
            check_row_order=False,
        )



#!/usr/bin/env python
"""Utilities for grouping and/or reducing MEDS cohort files by code to collect metadata properties."""

import logging
import time
from collections.abc import Callable, Sequence
from datetime import datetime
from enum import StrEnum
from pathlib import Path
from typing import NamedTuple

import hydra
import polars as pl
import polars.selectors as cs
from meds import subject_id_field
from omegaconf import DictConfig, ListConfig, OmegaConf

from MEDS_transforms import PREPROCESS_CONFIG_YAML
from MEDS_transforms.mapreduce.mapper import map_over
from MEDS_transforms.mapreduce.utils import is_complete_parquet_file
from MEDS_transforms.utils import write_lazyframe

logger = logging.getLogger(__name__)


class METADATA_FN(StrEnum):
    """Enumeration of metadata functions that can be applied to a group of codes.

    This enumeration contains the supported code-metadata collection and aggregation function names that can
    be applied to codes (or, rather, unique code & modifier units) in a MEDS cohort. Each function name is
    mapped, in the below `CODE_METADATA_AGGREGATIONS` dictionary, to mapper and reducer functions that (a)
    collect the raw data at a per code-modifier level from MEDS subject-level shards and (b) aggregates two or
    more per-shard metadata files into a single metadata file, which can be used to merge metadata across all
    shards into a single file.

    Note that, by design, these aggregations are all those that permit simple, single-variable reductions.
    E.g., rather than tracking the mean and standard deviation of a numeric value, we track the sum of the
    values and the sum of the squares of the values. This is because the mean and standard deviation can be
    trivially calculated from these two values, but ``sum`` and ``sum_sqd`` both can be reduced with a simple
    summation operation across shards, whereas the mean and standard deviation would require a more complex
    reduction function. This convention is _not_ motivated by any performance concerns -- such implications
    would be negligible -- but rather by a desire to keep the aggregation functions simple and easy to
    understand and verify, and by the reality that it is highly unlikely that one would ever care about the
    mean and standard deviation of a value on a per-shard, rather than per-dataset basis, for example, so the
    extra calculation being saved for after the full reduction process does not impact usability.

    These are stored as a `StrEnum` so that they can be easily specified by the user in a configuration file
    or on the command line.

    Args:
        "code/n_subjects": Collects the number of unique subjects who have (anywhere in their record) the code
            & modifiers group.
        "code/n_occurrences": Collects the total number of occurrences of the code & modifiers group across
            all observations for all subjects.
        "values/n_subjects": Collects the number of unique subjects who have a non-null, non-nan
            numeric_value field for the code & modifiers group.
        "values/n_occurrences": Collects the total number of non-null, non-nan numeric_value occurrences for
            the code & modifiers group across all observations for all subjects.
        "values/n_ints": Collects the number of times the observed, non-null numeric_value for the code &
            modifiers group is an integral value (i.e., a whole number, not an integral type).
        "values/sum": Collects the sum of the non-null, non-nan numeric_value values for the code &
            modifiers group.
        "values/sum_sqd": Collects the sum of the squares of the non-null, non-nan numeric_value values for
            the code
        "values/min": Collects the minimum non-null, non-nan numeric_value value for the code & modifiers
        "values/max": Collects the maximum non-null, non-nan numeric_value value for the code & modifiers
        "values/quantiles": Collects the specified quantiles over all observed numeric values for the code &
            modifiers group. The quantiles are specified in the output as a polars struct field, with the
            quantile as the key and the value as the quantile value. The desired quantiles are specified in
            the configuration file using the dictionary syntax for the aggregation.
    """

    CODE_N_PATIENTS = "code/n_subjects"
    CODE_N_OCCURRENCES = "code/n_occurrences"
    VALUES_N_PATIENTS = "values/n_subjects"
    VALUES_N_OCCURRENCES = "values/n_occurrences"
    VALUES_N_INTS = "values/n_ints"
    VALUES_SUM = "values/sum"
    VALUES_SUM_SQD = "values/sum_sqd"
    VALUES_MIN = "values/min"
    VALUES_MAX = "values/max"
    VALUES_QUANTILES = "values/quantiles"


class MapReducePair(NamedTuple):
    """A named tuple pair of a mapper and reducer function for type safety and clarity.

    The mapper (element 0) should be a polars expression that aggregates the data for a single metadata
    aggregation function. It must be suitable for use in a group_by operation on a polars LazyFrame. The
    reducer (element 1) should be a function that takes either a single polars expression, a sequence of
    polars expressions, or a polars Selector and reduces them into a single polars expression, which is
    returned. Built in functions such as `pl.sum_horizontal` and `pl.max_horizontal` are examples of functions
    that can be used as reducers. Custom defined functions can also be used.

    There is no validation of these criteria in the constructor, so it is up to the user to ensure that the
    functions provided are suitable for their intended use.

    Args:
        mapper: A polars expression that aggregates the data for a single metadata aggregation function. It
            must be suitable for use in a group_by operation on a polars LazyFrame.
        reducer: A function that takes either a single polars expression, a sequence of polars expressions, or
            a polars Selector and reduces them into a single polars expression, which is returned.
    """

    mapper: pl.Expr
    reducer: Callable[[pl.Expr | Sequence[pl.Expr] | cs._selector_proxy_], pl.Expr]


def quantile_reducer(cols: cs._selector_proxy_, quantiles: list[float]) -> pl.Expr:
    """Calculates the specified quantiles for the combined set of all numerical values in `cols`.

    Args:
        cols: A polars selector that selects the column(s) containing the numerical values for which the
            quantiles should be calculated.
        quantiles: A list of floats specifying the quantiles that should be calculated.

    Returns:
        A polars expression that calculates the specified quantiles for the combined set of all numerical
        values in `cols`.

    Examples:
        >>> df = pl.DataFrame({
        ...     "key": [1, 2],
        ...     "vals/shard1": [[1, 2, float('nan')], [None, 3]],
        ...     "vals/shard2": [[3.0, 4], [30]],
        ... }, strict=False)
        >>> expr = quantile_reducer(cs.starts_with("vals/"), [0.01, 0.5, 0.75])
        >>> df.select(expr)
        shape: (1, 1)
        ┌──────────────────┐
        │ values/quantiles │
        │ ---              │
        │ struct[3]        │
        ╞══════════════════╡
        │ {1.0,3.0,30.0}   │
        └──────────────────┘
        >>> df.select("key", expr.over("key"))
        shape: (2, 2)
        ┌─────┬──────────────────┐
        │ key ┆ values/quantiles │
        │ --- ┆ ---              │
        │ i64 ┆ struct[3]        │
        ╞═════╪══════════════════╡
        │ 1   ┆ {1.0,3.0,4.0}    │
        │ 2   ┆ {3.0,30.0,30.0}  │
        └─────┴──────────────────┘
    """

    vals = pl.concat_list(cols.fill_null([])).explode()

    quantile_cols = [f"values/quantile/{q}" for q in quantiles]
    quantiles_struct = {col: vals.quantile(q).alias(col) for col, q in zip(quantile_cols, quantiles)}

    return pl.struct(**quantiles_struct).alias(METADATA_FN.VALUES_QUANTILES)


VAL = pl.col("numeric_value")
VAL_PRESENT: pl.Expr = VAL.is_not_null() & VAL.is_not_nan()
IS_INT: pl.Expr = VAL.round() == VAL
PRESENT_VALS = VAL.filter(VAL_PRESENT)

CODE_METADATA_AGGREGATIONS: dict[METADATA_FN, MapReducePair] = {
    METADATA_FN.CODE_N_PATIENTS: MapReducePair(pl.col(subject_id_field).n_unique(), pl.sum_horizontal),
    METADATA_FN.CODE_N_OCCURRENCES: MapReducePair(pl.len(), pl.sum_horizontal),
    METADATA_FN.VALUES_N_PATIENTS: MapReducePair(
        pl.col(subject_id_field).filter(VAL_PRESENT).n_unique(), pl.sum_horizontal
    ),
    METADATA_FN.VALUES_N_OCCURRENCES: MapReducePair(PRESENT_VALS.len(), pl.sum_horizontal),
    METADATA_FN.VALUES_N_INTS: MapReducePair(VAL.filter(VAL_PRESENT & IS_INT).len(), pl.sum_horizontal),
    METADATA_FN.VALUES_SUM: MapReducePair(PRESENT_VALS.sum(), pl.sum_horizontal),
    METADATA_FN.VALUES_SUM_SQD: MapReducePair((PRESENT_VALS**2).sum(), pl.sum_horizontal),
    METADATA_FN.VALUES_MIN: MapReducePair(PRESENT_VALS.min(), pl.min_horizontal),
    METADATA_FN.VALUES_MAX: MapReducePair(PRESENT_VALS.max(), pl.max_horizontal),
    METADATA_FN.VALUES_QUANTILES: MapReducePair(PRESENT_VALS, quantile_reducer),
}


def validate_args_and_get_code_cols(stage_cfg: DictConfig, code_modifiers: list[str] | None) -> list[str]:
    """Validates the stage configuration and code_modifiers argument and returns the code group keys.

    Args:
        stage_cfg: The configuration object for this stage. It must contain an `aggregations` field that has a
            list of aggregations that should be applied in this stage. Each aggregation must be a string in
            the `METADATA_FN` enumeration.
        code_modifiers: A list of column names that should be used in addition to the core `code`
            column to group the data before applying the aggregations. If None, only the `code` column will be
            used.

    Returns:
        A list of column names that should be used to group the data before applying the aggregations.

    Raises:
        ValueError: If the stage config either does not contain an aggregations field, contains an empty or
            mis-typed aggregations field, or contains an invalid aggregation function.
        ValueError: If the code_modifiers argument is not a list of strings or None.

    Examples:
        >>> no_aggs_cfg = DictConfig({"other_key": "other_value"})
        >>> validate_args_and_get_code_cols(no_aggs_cfg, None) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Stage config must contain an 'aggregations' field. Got:
            other_key: other_value
        >>> invalid_agg_cfg = DictConfig({"aggregations": ["INVALID"]})
        >>> validate_args_and_get_code_cols(invalid_agg_cfg, None) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Metadata aggregation function INVALID not found in METADATA_FN enumeration. Values are:
            code/n_subjects, code/n_occurrences, values/n_subjects, values/n_occurrences, values/n_ints,
            values/sum, values/sum_sqd, values/min, values/max, values/quantiles
        >>> valid_cfg = DictConfig({"aggregations": ["code/n_subjects", {"name": "values/n_ints"}]})
        >>> validate_args_and_get_code_cols(valid_cfg, 33)
        Traceback (most recent call last):
            ...
        ValueError: code_modifiers must be a list of strings or None. Got 33
        >>> validate_args_and_get_code_cols(valid_cfg, [33])
        Traceback (most recent call last):
            ...
        ValueError: code_modifiers must be a list of strings or None. Got [33]
        >>> validate_args_and_get_code_cols(valid_cfg, ["modifier1"])
        ['code', 'modifier1']
        >>> validate_args_and_get_code_cols(valid_cfg, None)
        ['code']
    """

    if "aggregations" not in stage_cfg:
        raise ValueError(
            f"Stage config must contain an 'aggregations' field. Got:\n{OmegaConf.to_yaml(stage_cfg)}"
        )

    aggregations = stage_cfg.aggregations
    for agg in aggregations:
        if isinstance(agg, (dict, DictConfig)):
            agg = agg.get("name", None)
        if agg not in {fn.value for fn in METADATA_FN}:
            raise ValueError(
                f"Metadata aggregation function {agg} not found in METADATA_FN enumeration. Values are: "
                f"{', '.join([fn.value for fn in METADATA_FN])}"
            )

    match code_modifiers:
        case None:
            return ["code"]
        case list() | ListConfig() if all(isinstance(col, str) for col in code_modifiers):
            return ["code"] + code_modifiers
        case _:
            raise ValueError(f"code_modifiers must be a list of strings or None. Got {code_modifiers}")


def mapper_fntr(
    stage_cfg: DictConfig, code_modifiers: list[str] | None
) -> Callable[[pl.DataFrame], pl.DataFrame]:
    """Returns a function that extracts code metadata from a MEDS cohort shard.

    Args:
        stage_cfg: The configuration object for this stage. It must contain an `aggregations` field that has a
            list of aggregations that should be applied in this stage. Each aggregation must be a string in
            the `METADATA_FN` enumeration, and the mapper function is specified in the
            `CODE_METADATA_AGGREGATIONS` dictionary.
        code_modifiers: A list of column names that should be used in addition to the core `code`
            column to group the data before applying the aggregations. If None, only the `code` column will be
            used.

    Raises: See `validate_args_and_get_code_cols`.

    Returns:
        A function that extracts the specified metadata from a MEDS cohort shard after grouping by the
        specified code & modifier columns. **Note**: The output of this function will, if
        ``stage_cfg.do_summarize_over_all_codes`` is True, contain the metadata summarizing all observations
        across all codes and subjects in the shard, with both ``code`` and all ``code_modifiers`` set
        to `None` in the output dataframe, in the same format as the code/modifier specific rows with non-null
        values.

    Examples:
        >>> import numpy as np
        >>> df = pl.DataFrame({
        ...     "code":             ["A", "B", "A", "B", "C", "A", "C", "B",          "D"],
        ...     "modifier1":        [1,   2,   1,   2,   1,   2,   1,   2,            None],
        ...     "modifier_ignored": [3,   3,   4,   4,   5,   5,   6,   6,            7],
        ...     "subject_id":       [1,   2,   1,   3,   1,   2,   2,   2,            1],
        ...     "numeric_value":    [1.1, 2.,  1.1, 4.,  5.,  6.,  7.5, float('nan'), None],
        ... })
        >>> df
        shape: (9, 5)
        ┌──────┬───────────┬──────────────────┬────────────┬───────────────┐
        │ code ┆ modifier1 ┆ modifier_ignored ┆ subject_id ┆ numeric_value │
        │ ---  ┆ ---       ┆ ---              ┆ ---        ┆ ---           │
        │ str  ┆ i64       ┆ i64              ┆ i64        ┆ f64           │
        ╞══════╪═══════════╪══════════════════╪════════════╪═══════════════╡
        │ A    ┆ 1         ┆ 3                ┆ 1          ┆ 1.1           │
        │ B    ┆ 2         ┆ 3                ┆ 2          ┆ 2.0           │
        │ A    ┆ 1         ┆ 4                ┆ 1          ┆ 1.1           │
        │ B    ┆ 2         ┆ 4                ┆ 3          ┆ 4.0           │
        │ C    ┆ 1         ┆ 5                ┆ 1          ┆ 5.0           │
        │ A    ┆ 2         ┆ 5                ┆ 2          ┆ 6.0           │
        │ C    ┆ 1         ┆ 6                ┆ 2          ┆ 7.5           │
        │ B    ┆ 2         ┆ 6                ┆ 2          ┆ NaN           │
        │ D    ┆ null      ┆ 7                ┆ 1          ┆ null          │
        └──────┴───────────┴──────────────────┴────────────┴───────────────┘
        >>> stage_cfg = DictConfig({
        ...     "aggregations": ["code/n_subjects", "values/n_ints"],
        ...     "do_summarize_over_all_codes": True
        ... })
        >>> mapper = mapper_fntr(stage_cfg, None)
        >>> mapper(df.lazy()).collect()
        shape: (5, 3)
        ┌──────┬─────────────────┬───────────────┐
        │ code ┆ code/n_subjects ┆ values/n_ints │
        │ ---  ┆ ---             ┆ ---           │
        │ str  ┆ u32             ┆ u32           │
        ╞══════╪═════════════════╪═══════════════╡
        │ null ┆ 3               ┆ 4             │
        │ A    ┆ 2               ┆ 1             │
        │ B    ┆ 2               ┆ 2             │
        │ C    ┆ 2               ┆ 1             │
        │ D    ┆ 1               ┆ 0             │
        └──────┴─────────────────┴───────────────┘
        >>> stage_cfg = DictConfig({"aggregations": ["code/n_subjects", "values/n_ints"]})
        >>> mapper = mapper_fntr(stage_cfg, None)
        >>> mapper(df.lazy()).collect()
        shape: (4, 3)
        ┌──────┬─────────────────┬───────────────┐
        │ code ┆ code/n_subjects ┆ values/n_ints │
        │ ---  ┆ ---             ┆ ---           │
        │ str  ┆ u32             ┆ u32           │
        ╞══════╪═════════════════╪═══════════════╡
        │ A    ┆ 2               ┆ 1             │
        │ B    ┆ 2               ┆ 2             │
        │ C    ┆ 2               ┆ 1             │
        │ D    ┆ 1               ┆ 0             │
        └──────┴─────────────────┴───────────────┘
        >>> code_modifiers = ["modifier1"]
        >>> stage_cfg = DictConfig({"aggregations": ["code/n_subjects", "values/n_ints"]})
        >>> mapper = mapper_fntr(stage_cfg, ListConfig(code_modifiers))
        >>> mapper(df.lazy()).collect()
        shape: (5, 4)
        ┌──────┬───────────┬─────────────────┬───────────────┐
        │ code ┆ modifier1 ┆ code/n_subjects ┆ values/n_ints │
        │ ---  ┆ ---       ┆ ---             ┆ ---           │
        │ str  ┆ i64       ┆ u32             ┆ u32           │
        ╞══════╪═══════════╪═════════════════╪═══════════════╡
        │ A    ┆ 1         ┆ 1               ┆ 0             │
        │ A    ┆ 2         ┆ 1               ┆ 1             │
        │ B    ┆ 2         ┆ 2               ┆ 2             │
        │ C    ┆ 1         ┆ 2               ┆ 1             │
        │ D    ┆ null      ┆ 1               ┆ 0             │
        └──────┴───────────┴─────────────────┴───────────────┘
        >>> stage_cfg = DictConfig({"aggregations": ["code/n_occurrences", "values/sum"]})
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df.lazy()).collect()
        shape: (5, 4)
        ┌──────┬───────────┬────────────────────┬────────────┐
        │ code ┆ modifier1 ┆ code/n_occurrences ┆ values/sum │
        │ ---  ┆ ---       ┆ ---                ┆ ---        │
        │ str  ┆ i64       ┆ u32                ┆ f64        │
        ╞══════╪═══════════╪════════════════════╪════════════╡
        │ A    ┆ 1         ┆ 2                  ┆ 2.2        │
        │ A    ┆ 2         ┆ 1                  ┆ 6.0        │
        │ B    ┆ 2         ┆ 3                  ┆ 6.0        │
        │ C    ┆ 1         ┆ 2                  ┆ 12.5       │
        │ D    ┆ null      ┆ 1                  ┆ 0.0        │
        └──────┴───────────┴────────────────────┴────────────┘
        >>> stage_cfg = DictConfig({
        ...     "aggregations": ["code/n_occurrences", "values/sum"],
        ...     "do_summarize_over_all_codes": True,
        ... })
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df.lazy()).collect()
        shape: (6, 4)
        ┌──────┬───────────┬────────────────────┬────────────┐
        │ code ┆ modifier1 ┆ code/n_occurrences ┆ values/sum │
        │ ---  ┆ ---       ┆ ---                ┆ ---        │
        │ str  ┆ i64       ┆ u32                ┆ f64        │
        ╞══════╪═══════════╪════════════════════╪════════════╡
        │ null ┆ null      ┆ 9                  ┆ 26.7       │
        │ A    ┆ 1         ┆ 2                  ┆ 2.2        │
        │ A    ┆ 2         ┆ 1                  ┆ 6.0        │
        │ B    ┆ 2         ┆ 3                  ┆ 6.0        │
        │ C    ┆ 1         ┆ 2                  ┆ 12.5       │
        │ D    ┆ null      ┆ 1                  ┆ 0.0        │
        └──────┴───────────┴────────────────────┴────────────┘
        >>> stage_cfg = DictConfig({"aggregations": ["values/n_subjects", "values/n_occurrences"]})
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df.lazy()).collect()
        shape: (5, 4)
        ┌──────┬───────────┬───────────────────┬──────────────────────┐
        │ code ┆ modifier1 ┆ values/n_subjects ┆ values/n_occurrences │
        │ ---  ┆ ---       ┆ ---               ┆ ---                  │
        │ str  ┆ i64       ┆ u32               ┆ u32                  │
        ╞══════╪═══════════╪═══════════════════╪══════════════════════╡
        │ A    ┆ 1         ┆ 1                 ┆ 2                    │
        │ A    ┆ 2         ┆ 1                 ┆ 1                    │
        │ B    ┆ 2         ┆ 2                 ┆ 2                    │
        │ C    ┆ 1         ┆ 2                 ┆ 2                    │
        │ D    ┆ null      ┆ 0                 ┆ 0                    │
        └──────┴───────────┴───────────────────┴──────────────────────┘
        >>> stage_cfg = DictConfig({"aggregations": ["values/sum_sqd", "values/min", "values/max"]})
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df.lazy()).collect()
        shape: (5, 5)
        ┌──────┬───────────┬────────────────┬────────────┬────────────┐
        │ code ┆ modifier1 ┆ values/sum_sqd ┆ values/min ┆ values/max │
        │ ---  ┆ ---       ┆ ---            ┆ ---        ┆ ---        │
        │ str  ┆ i64       ┆ f64            ┆ f64        ┆ f64        │
        ╞══════╪═══════════╪════════════════╪════════════╪════════════╡
        │ A    ┆ 1         ┆ 2.42           ┆ 1.1        ┆ 1.1        │
        │ A    ┆ 2         ┆ 36.0           ┆ 6.0        ┆ 6.0        │
        │ B    ┆ 2         ┆ 20.0           ┆ 2.0        ┆ 4.0        │
        │ C    ┆ 1         ┆ 81.25          ┆ 5.0        ┆ 7.5        │
        │ D    ┆ null      ┆ 0.0            ┆ null       ┆ null       │
        └──────┴───────────┴────────────────┴────────────┴────────────┘
        >>> stage_cfg = DictConfig({
        ...     "aggregations": [{"name": "values/quantiles", "quantiles": [0.25, 0.5, 0.75]}]
        ... })
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df.lazy()).collect().select("code", "modifier1", pl.col("values/quantiles"))
        shape: (5, 3)
        ┌──────┬───────────┬──────────────────┐
        │ code ┆ modifier1 ┆ values/quantiles │
        │ ---  ┆ ---       ┆ ---              │
        │ str  ┆ i64       ┆ list[f64]        │
        ╞══════╪═══════════╪══════════════════╡
        │ A    ┆ 1         ┆ [1.1, 1.1]       │
        │ A    ┆ 2         ┆ [6.0]            │
        │ B    ┆ 2         ┆ [2.0, 4.0]       │
        │ C    ┆ 1         ┆ [5.0, 7.5]       │
        │ D    ┆ null      ┆ []               │
        └──────┴───────────┴──────────────────┘
        >>> stage_cfg = DictConfig({
        ...     "aggregations": [{"name": "values/quantiles", "quantiles": [0.25, 0.5, 0.75]}],
        ...     "do_summarize_over_all_codes": True,
        ... })
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df.lazy()).collect().select("code", "modifier1", pl.col("values/quantiles"))
        shape: (6, 3)
        ┌──────┬───────────┬───────────────────┐
        │ code ┆ modifier1 ┆ values/quantiles  │
        │ ---  ┆ ---       ┆ ---               │
        │ str  ┆ i64       ┆ list[f64]         │
        ╞══════╪═══════════╪═══════════════════╡
        │ null ┆ null      ┆ [1.1, 2.0, … 7.5] │
        │ A    ┆ 1         ┆ [1.1, 1.1]        │
        │ A    ┆ 2         ┆ [6.0]             │
        │ B    ┆ 2         ┆ [2.0, 4.0]        │
        │ C    ┆ 1         ┆ [5.0, 7.5]        │
        │ D    ┆ null      ┆ []                │
        └──────┴───────────┴───────────────────┘

    Empty dataframes are handled as you would expect
        >>> df_empty = pl.DataFrame({
        ...     "code": [],
        ...     "modifier1": [],
        ...     "modifier_ignored": [],
        ...     "subject_id": [],
        ...     "numeric_value": [],
        ... }, schema=df.schema)
        >>> stage_cfg = DictConfig({"aggregations": ["values/sum_sqd", "values/min", "values/max"]})
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df_empty.lazy()).collect()
        shape: (0, 5)
        ┌──────┬───────────┬────────────────┬────────────┬────────────┐
        │ code ┆ modifier1 ┆ values/sum_sqd ┆ values/min ┆ values/max │
        │ ---  ┆ ---       ┆ ---            ┆ ---        ┆ ---        │
        │ str  ┆ i64       ┆ f64            ┆ f64        ┆ f64        │
        ╞══════╪═══════════╪════════════════╪════════════╪════════════╡
        └──────┴───────────┴────────────────┴────────────┴────────────┘
        >>> stage_cfg = DictConfig({
        ...     "aggregations": ["values/sum_sqd", "values/min", "values/max"],
        ...     "do_summarize_over_all_codes": True,
        ... })
        >>> mapper = mapper_fntr(stage_cfg, code_modifiers)
        >>> mapper(df_empty.lazy()).collect()
        shape: (1, 5)
        ┌──────┬───────────┬────────────────┬────────────┬────────────┐
        │ code ┆ modifier1 ┆ values/sum_sqd ┆ values/min ┆ values/max │
        │ ---  ┆ ---       ┆ ---            ┆ ---        ┆ ---        │
        │ str  ┆ i64       ┆ f64            ┆ f64        ┆ f64        │
        ╞══════╪═══════════╪════════════════╪════════════╪════════════╡
        │ null ┆ null      ┆ 0.0            ┆ null       ┆ null       │
        └──────┴───────────┴────────────────┴────────────┴────────────┘
    """

    code_key_columns = validate_args_and_get_code_cols(stage_cfg, code_modifiers)
    aggregations = stage_cfg.aggregations

    agg_operations = {}
    for agg in aggregations:
        agg_name = agg if isinstance(agg, str) else agg["name"]
        agg_operations[agg_name] = CODE_METADATA_AGGREGATIONS[agg_name].mapper

    def by_code_mapper(df: pl.LazyFrame) -> pl.LazyFrame:
        return df.group_by(code_key_columns).agg(**agg_operations).sort(code_key_columns)

    def all_subjects_mapper(df: pl.LazyFrame) -> pl.LazyFrame:
        local_agg_operations = agg_operations.copy()
        if METADATA_FN.VALUES_QUANTILES in agg_operations:
            local_agg_operations[METADATA_FN.VALUES_QUANTILES] = agg_operations[
                METADATA_FN.VALUES_QUANTILES
            ].implode()
        return df.select(**local_agg_operations)

    if stage_cfg.get("do_summarize_over_all_codes", False):

        def mapper(df: pl.LazyFrame) -> pl.LazyFrame:
            by_code = by_code_mapper(df)
            all_subjects = all_subjects_mapper(df)
            return pl.concat([all_subjects, by_code], how="diagonal_relaxed").select(
                *code_key_columns, *agg_operations.keys()
            )

    else:
        mapper = by_code_mapper

    return mapper


def reducer_fntr(
    stage_cfg: DictConfig, code_modifiers: list[str] | None = None
) -> Callable[[Sequence[pl.DataFrame]], pl.DataFrame]:
    """Returns a function that merges different code metadata files together into an aggregated total.

    Args:
        stage_cfg: The configuration object for this stage. It must contain an `aggregations` field that has a
            list of aggregations that should be applied in this stage. Each aggregation must be a string in
            the `METADATA_FN` enumeration, and the reduction function is specified in the
            `CODE_METADATA_AGGREGATIONS` dictionary.
        code_modifiers: A list of column names that should be used in addition to the core `code`
            column to group the data before applying the aggregations. If None, only the `code` column will be
            used.

    Returns:
        A function that aggregates the specified metadata columns from different extracted metadata shards
        into a total view.

    Raises: See `validate_args_and_get_code_cols`.

    Examples:
        >>> df_1 = pl.DataFrame({
        ...     "code": [None, "A", "A", "B", "C"],
        ...     "modifier1": [None, 1, 2, 1, 2],
        ...     "code/n_subjects":  [10, 1, 1, 2, 2],
        ...     "code/n_occurrences": [13, 2, 1, 3, 2],
        ...     "values/n_subjects":  [8, 1, 1, 2, 2],
        ...     "values/n_occurrences": [12, 2, 1, 3, 2],
        ...     "values/n_ints": [4, 0, 1, 3, 1],
        ...     "values/sum": [13.2, 2.2, 6.0, 14.0, 12.5],
        ...     "values/sum_sqd": [21.3, 2.42, 36.0, 84.0, 81.25],
        ...     "values/min": [-1, 0, -1, 2, 2],
        ...     "values/max": [8.0, 1.1, 6.0, 8.0, 7.5],
        ...     "values/quantiles": [[1.1, 1.1], [6.0], [6.0], [5.0, 7.5], []],
        ... })
        >>> df_2 = pl.DataFrame({
        ...     "code": ["A", "A", "B", "C"],
        ...     "modifier1": [1, 2, 1, None],
        ...     "code/n_subjects":  [3, 3, 4, 4],
        ...     "code/n_occurrences": [10, 11, 8, 11],
        ...     "values/n_subjects":  [0, 1, 2, 2],
        ...     "values/n_occurrences": [0, 4, 3, 2],
        ...     "values/n_ints": [0, 1, 3, 1],
        ...     "values/sum": [0., 7.0, 14.0, 12.5],
        ...     "values/sum_sqd": [0., 103.2, 84.0, 81.25],
        ...     "values/min": [None, -1., 0.2, -2.],
        ...     "values/max": [None, 6.2, 1.0, 1.5],
        ...     "values/quantiles": [[1.3, -1.1, 2.0], [6.0, 1.2], [3.0, 2.5], [11.1, 12.]],
        ... })
        >>> df_3 = pl.DataFrame({
        ...     "code": ["D"],
        ...     "modifier1": [1],
        ...     "code/n_subjects": [2],
        ...     "code/n_occurrences": [2],
        ...     "values/n_subjects": [1],
        ...     "values/n_occurrences": [3],
        ...     "values/n_ints": [3],
        ...     "values/sum": [2],
        ...     "values/sum_sqd": [4],
        ...     "values/min": [0],
        ...     "values/max": [2],
        ...     "values/quantiles": [[]],
        ... })
        >>> df_empty = pl.DataFrame({
        ...     "code": [],
        ...     "modifier1": [],
        ...     "code/n_subjects": [],
        ...     "code/n_occurrences": [],
        ...     "values/n_subjects": [],
        ...     "values/n_occurrences": [],
        ...     "values/n_ints": [],
        ...     "values/sum": [],
        ...     "values/sum_sqd": [],
        ...     "values/min": [],
        ...     "values/max": [],
        ...     "values/quantiles": [],
        ... }, schema=df_3.schema)
        >>> df_null_empty = pl.DataFrame({
        ...     "code": [None],
        ...     "modifier1": [None],
        ...     "code/n_subjects": [0],
        ...     "code/n_occurrences": [0],
        ...     "values/n_subjects": [0],
        ...     "values/n_occurrences": [0],
        ...     "values/n_ints": [0],
        ...     "values/sum": [0],
        ...     "values/sum_sqd": [0],
        ...     "values/min": [None],
        ...     "values/max": [None],
        ...     "values/quantiles": [None],
        ... }, schema=df_3.schema)
        >>> code_modifiers = ["modifier1"]
        >>> stage_cfg = DictConfig({"aggregations": ["code/n_subjects", "values/n_ints"]})
        >>> reducer = reducer_fntr(stage_cfg, code_modifiers)
        >>> reducer(df_1, df_2, df_3, df_empty, df_null_empty)
        shape: (7, 4)
        ┌──────┬───────────┬─────────────────┬───────────────┐
        │ code ┆ modifier1 ┆ code/n_subjects ┆ values/n_ints │
        │ ---  ┆ ---       ┆ ---             ┆ ---           │
        │ str  ┆ i64       ┆ i64             ┆ i64           │
        ╞══════╪═══════════╪═════════════════╪═══════════════╡
        │ null ┆ null      ┆ 10              ┆ 4             │
        │ A    ┆ 1         ┆ 4               ┆ 0             │
        │ A    ┆ 2         ┆ 4               ┆ 2             │
        │ B    ┆ 1         ┆ 6               ┆ 6             │
        │ C    ┆ null      ┆ 4               ┆ 1             │
        │ C    ┆ 2         ┆ 2               ┆ 1             │
        │ D    ┆ 1         ┆ 2               ┆ 3             │
        └──────┴───────────┴─────────────────┴───────────────┘
        >>> cfg = DictConfig({
        ...     "code_modifiers": ["modifier1"],
        ...     "code_processing_stages": {
        ...         "stage1": ["code/n_subjects", "values/n_ints"],
        ...         "stage2": ["code/n_occurrences", "values/sum"],
        ...         "stage3.A": ["values/n_subjects", "values/n_occurrences"],
        ...         "stage3.B": ["values/sum_sqd", "values/min", "values/max"],
        ...         "stage4": ["INVALID"],
        ...     }
        ... })
        >>> stage_cfg = DictConfig({"aggregations": ["code/n_occurrences", "values/sum"]})
        >>> reducer = reducer_fntr(stage_cfg, code_modifiers)
        >>> reducer(df_1, df_2, df_3, df_empty, df_null_empty)
        shape: (7, 4)
        ┌──────┬───────────┬────────────────────┬────────────┐
        │ code ┆ modifier1 ┆ code/n_occurrences ┆ values/sum │
        │ ---  ┆ ---       ┆ ---                ┆ ---        │
        │ str  ┆ i64       ┆ i64                ┆ f64        │
        ╞══════╪═══════════╪════════════════════╪════════════╡
        │ null ┆ null      ┆ 13                 ┆ 13.2       │
        │ A    ┆ 1         ┆ 12                 ┆ 2.2        │
        │ A    ┆ 2         ┆ 12                 ┆ 13.0       │
        │ B    ┆ 1         ┆ 11                 ┆ 28.0       │
        │ C    ┆ null      ┆ 11                 ┆ 12.5       │
        │ C    ┆ 2         ┆ 2                  ┆ 12.5       │
        │ D    ┆ 1         ┆ 2                  ┆ 2.0        │
        └──────┴───────────┴────────────────────┴────────────┘
        >>> stage_cfg = DictConfig({"aggregations": ["values/n_subjects", "values/n_occurrences"]})
        >>> reducer = reducer_fntr(stage_cfg, code_modifiers)
        >>> reducer(df_1, df_2, df_3, df_empty, df_null_empty)
        shape: (7, 4)
        ┌──────┬───────────┬───────────────────┬──────────────────────┐
        │ code ┆ modifier1 ┆ values/n_subjects ┆ values/n_occurrences │
        │ ---  ┆ ---       ┆ ---               ┆ ---                  │
        │ str  ┆ i64       ┆ i64               ┆ i64                  │
        ╞══════╪═══════════╪═══════════════════╪══════════════════════╡
        │ null ┆ null      ┆ 8                 ┆ 12                   │
        │ A    ┆ 1         ┆ 1                 ┆ 2                    │
        │ A    ┆ 2         ┆ 2                 ┆ 5                    │
        │ B    ┆ 1         ┆ 4                 ┆ 6                    │
        │ C    ┆ null      ┆ 2                 ┆ 2                    │
        │ C    ┆ 2         ┆ 2                 ┆ 2                    │
        │ D    ┆ 1         ┆ 1                 ┆ 3                    │
        └──────┴───────────┴───────────────────┴──────────────────────┘
        >>> stage_cfg = DictConfig({"aggregations": ["values/sum_sqd", "values/min", "values/max"]})
        >>> reducer = reducer_fntr(stage_cfg, code_modifiers)
        >>> reducer(df_1, df_2, df_3)
        shape: (7, 5)
        ┌──────┬───────────┬────────────────┬────────────┬────────────┐
        │ code ┆ modifier1 ┆ values/sum_sqd ┆ values/min ┆ values/max │
        │ ---  ┆ ---       ┆ ---            ┆ ---        ┆ ---        │
        │ str  ┆ i64       ┆ f64            ┆ f64        ┆ f64        │
        ╞══════╪═══════════╪════════════════╪════════════╪════════════╡
        │ null ┆ null      ┆ 21.3           ┆ -1.0       ┆ 8.0        │
        │ A    ┆ 1         ┆ 2.42           ┆ 0.0        ┆ 1.1        │
        │ A    ┆ 2         ┆ 139.2          ┆ -1.0       ┆ 6.2        │
        │ B    ┆ 1         ┆ 168.0          ┆ 0.2        ┆ 8.0        │
        │ C    ┆ null      ┆ 81.25          ┆ -2.0       ┆ 1.5        │
        │ C    ┆ 2         ┆ 81.25          ┆ 2.0        ┆ 7.5        │
        │ D    ┆ 1         ┆ 4.0            ┆ 0.0        ┆ 2.0        │
        └──────┴───────────┴────────────────┴────────────┴────────────┘
        >>> reducer(df_1.drop("values/min"), df_2, df_3)
        Traceback (most recent call last):
            ...
        KeyError: 'Column values/min not found in DataFrame 0 for reduction.'
        >>> stage_cfg = DictConfig({
        ...     "aggregations": [{"name": "values/quantiles", "quantiles": [0.25, 0.5, 0.75]}],
        ... })
        >>> reducer = reducer_fntr(stage_cfg, code_modifiers)
        >>> reducer(df_1, df_2, df_3, df_empty, df_null_empty).unnest("values/quantiles")
        shape: (7, 5)
        ┌──────┬───────────┬──────────────────────┬─────────────────────┬──────────────────────┐
        │ code ┆ modifier1 ┆ values/quantile/0.25 ┆ values/quantile/0.5 ┆ values/quantile/0.75 │
        │ ---  ┆ ---       ┆ ---                  ┆ ---                 ┆ ---                  │
        │ str  ┆ i64       ┆ f64                  ┆ f64                 ┆ f64                  │
        ╞══════╪═══════════╪══════════════════════╪═════════════════════╪══════════════════════╡
        │ null ┆ null      ┆ 1.1                  ┆ 1.1                 ┆ 1.1                  │
        │ A    ┆ 1         ┆ 1.3                  ┆ 2.0                 ┆ 2.0                  │
        │ A    ┆ 2         ┆ 6.0                  ┆ 6.0                 ┆ 6.0                  │
        │ B    ┆ 1         ┆ 3.0                  ┆ 5.0                 ┆ 5.0                  │
        │ C    ┆ null      ┆ 11.1                 ┆ 12.0                ┆ 12.0                 │
        │ C    ┆ 2         ┆ null                 ┆ null                ┆ null                 │
        │ D    ┆ 1         ┆ null                 ┆ null                ┆ null                 │
        └──────┴───────────┴──────────────────────┴─────────────────────┴──────────────────────┘
    """

    code_key_columns = validate_args_and_get_code_cols(stage_cfg, code_modifiers)
    aggregations = stage_cfg.aggregations

    agg_operations = {}
    for agg in aggregations:
        if isinstance(agg, (dict, DictConfig)):
            agg_name = agg["name"]
            agg_kwargs = {k: v for k, v in agg.items() if k != "name"}
        else:
            agg_name = agg
            agg_kwargs = {}
        agg_operations[agg_name] = (
            CODE_METADATA_AGGREGATIONS[agg_name]
            .reducer(cs.matches(f"{agg_name}/shard_\\d+"), **agg_kwargs)
            .over(*code_key_columns)
        )

    def reducer(*dfs: Sequence[pl.LazyFrame]) -> pl.LazyFrame:
        renamed_dfs = []
        for i, df in enumerate(dfs):
            agg_selectors = []
            for agg in aggregations:
                if isinstance(agg, (dict, DictConfig)):
                    agg = agg["name"]
                if agg not in df.columns:
                    raise KeyError(f"Column {agg} not found in DataFrame {i} for reduction.")
                agg_selectors.append(pl.col(agg).alias(f"{agg}/shard_{i}"))

            renamed_dfs.append(df.select(*code_key_columns, *agg_selectors))

        df = renamed_dfs[0]
        for rdf in renamed_dfs[1:]:
            df = df.join(rdf, on=code_key_columns, how="full", join_nulls=True, coalesce=True)

        return df.select(*code_key_columns, **agg_operations).sort(code_key_columns)

    return reducer


def run_map_reduce(cfg: DictConfig):
    """Stored separately so it can be easily imported into the pre-built extraction pipelines."""
    all_out_fps = map_over(cfg, compute_fn=mapper_fntr)

    if cfg.worker != 0:
        logger.info("Code metadata mapping completed. Exiting")
        return

    logger.info("Starting reduction process")

    while not all(is_complete_parquet_file(fp) for fp in all_out_fps):  # pragma: no cover
        logger.info("Waiting to begin reduction for all files to be written...")
        time.sleep(cfg.polling_time)

    start = datetime.now()
    logger.info("All map shards complete! Starting code metadata reduction computation.")
    reducer_fp = Path(cfg.stage_cfg.reducer_output_dir) / "codes.parquet"
    reducer_fp.parent.mkdir(parents=True, exist_ok=True)

    reducer_fn = reducer_fntr(cfg.stage_cfg, cfg.get("code_modifiers", None))
    reduced = reducer_fn(*[pl.scan_parquet(fp, glob=False) for fp in all_out_fps]).with_columns(
        cs.numeric().shrink_dtype().name.keep()
    )

    old_metadata_fp = Path(cfg.stage_cfg.metadata_input_dir) / "codes.parquet"
    join_cols = ["code", *cfg.get("code_modifier_cols", [])]

    if old_metadata_fp.exists():
        logger.info(f"Joining to existing code metadata at {str(old_metadata_fp.resolve())}")
        existing = pl.scan_parquet(old_metadata_fp)
        existing = existing.drop(*[c for c in existing.columns if c in set(reduced.columns) - set(join_cols)])
        reduced = reduced.join(existing, on=join_cols, how="left", coalesce=True)

    write_lazyframe(reduced, reducer_fp)
    logger.info(f"Finished reduction in {datetime.now() - start}")


@hydra.main(
    version_base=None, config_path=str(PREPROCESS_CONFIG_YAML.parent), config_name=PREPROCESS_CONFIG_YAML.stem
)
def main(cfg: DictConfig):
    """Aggregates code metadata across shards.

    Note that the output of this stage includes a `null` code row if
    `stage_configs.STAGE_NAME.do_summarize_over_all_codes` is True. This row contains the total counts across
    all codes, _not_ the counts for rows with a code that is `null`, which should not happen.
    """

    run_map_reduce(cfg)


if __name__ == "__main__":  # pragma: no cover
    main()



"""Tests the finalize MEDS metadata process.

Set the bash env variable `DO_USE_LOCAL_SCRIPTS=1` to use the local py files, rather than the installed
scripts.
"""

from datetime import datetime

import polars as pl
from meds import __version__ as MEDS_VERSION
from meds import code_metadata_filepath, dataset_metadata_filepath, subject_splits_filepath

from MEDS_transforms.utils import get_package_version as get_meds_transform_version
from tests.MEDS_Extract import FINALIZE_METADATA_SCRIPT
from tests.utils import single_stage_tester

SHARDS_JSON = {
    "train/0": [239684, 1195293],
    "train/1": [68729, 814703],
    "tuning/0": [754281],
    "held_out/0": [1500733],
}

WANT_OUTPUTS = {
    "metadata/codes": pl.DataFrame(
        {
            "code": ["EYE_COLOR//BLUE", "EYE_COLOR//BROWN", "EYE_COLOR//HAZEL", "HR", "TEMP"],
            "description": [
                "Blue Eyes. Less common than brown.",
                "Brown Eyes. The most common eye color.",
                "Hazel eyes. These are uncommon",
                "Heart Rate",
                "Body Temperature",
            ],
            "parent_codes": [None, None, None, ["LOINC/8867-4"], ["LOINC/8310-5"]],
        }
    ),
}

METADATA_DF = pl.DataFrame(
    {
        "code": ["EYE_COLOR//BLUE", "EYE_COLOR//BROWN", "EYE_COLOR//HAZEL", "HR", "TEMP"],
        "description": [
            "Blue Eyes. Less common than brown.",
            "Brown Eyes. The most common eye color.",
            "Hazel eyes. These are uncommon",
            "Heart Rate",
            "Body Temperature",
        ],
        "parent_codes": [None, None, None, ["LOINC/8867-4"], ["LOINC/8310-5"]],
    }
)


def want_dataset_metadata(got: dict):
    want_known = {
        "dataset_name": "TEST",
        "dataset_version": "1.0",
        "etl_name": "MEDS_transforms",
        "etl_version": get_meds_transform_version(),
        "meds_version": MEDS_VERSION,
    }

    assert "created_at" in got, "Expected 'created_at' to be in the dataset metadata."
    created_at_obs = got.pop("created_at")
    as_dt = datetime.fromisoformat(created_at_obs)
    assert as_dt < datetime.now(), f"Expected 'created_at' to be before now, got {created_at_obs}."
    created_ago = datetime.now() - as_dt
    assert created_ago.total_seconds() < 5 * 60, "Expected 'created_at' to be within 5 minutes of now."

    assert got == want_known, f"Expected dataset metadata (less created at) to be {want_known}, got {got}."


WANT_OUTPUTS = {
    code_metadata_filepath: (
        METADATA_DF.with_columns(
            pl.col("code").cast(pl.String),
            pl.col("description").cast(pl.String),
            pl.col("parent_codes").cast(pl.List(pl.String)),
        ).select(["code", "description", "parent_codes"])
    ),
    subject_splits_filepath: pl.DataFrame(
        {
            "subject_id": [239684, 1195293, 68729, 814703, 754281, 1500733],
            "split": ["train", "train", "train", "train", "tuning", "held_out"],
        }
    ),
    dataset_metadata_filepath: want_dataset_metadata,
}


def test_convert_to_sharded_events():
    single_stage_tester(
        script=FINALIZE_METADATA_SCRIPT,
        stage_name="finalize_MEDS_metadata",
        stage_kwargs=None,
        config_name="extract",
        input_files={
            "metadata/codes": METADATA_DF,
            "metadata/.shards.json": SHARDS_JSON,
        },
        **{"etl_metadata.dataset_name": "TEST", "etl_metadata.dataset_version": "1.0"},
        event_conversion_config_fp="{input_dir}/event_cfgs.yaml",
        shards_map_fp="{input_dir}/metadata/.shards.json",
        want_outputs=WANT_OUTPUTS,
        df_check_kwargs={"check_row_order": False, "check_column_order": True, "check_dtypes": True},
    )



#!/usr/bin/env python
"""Utilities for extracting code metadata about the codes produced for the MEDS events."""

import copy
import logging
import random
import time
from datetime import datetime
from functools import partial
from pathlib import Path

import hydra
import polars as pl
from omegaconf import DictConfig, OmegaConf

logger = logging.getLogger(__name__)

from MEDS_transforms.extract import CONFIG_YAML, MEDS_METADATA_MANDATORY_TYPES
from MEDS_transforms.extract.convert_to_sharded_events import get_code_expr
from MEDS_transforms.extract.utils import get_supported_fp
from MEDS_transforms.mapreduce.mapper import rwlock_wrap
from MEDS_transforms.parser import cfg_to_expr
from MEDS_transforms.utils import stage_init, write_lazyframe


def extract_metadata(
    metadata_df: pl.LazyFrame, event_cfg: dict[str, str | None], allowed_codes: list | None = None
) -> pl.LazyFrame:
    """Extracts a single metadata dataframe block for an event configuration from the raw metadata.

    Args:
        df: The raw metadata DataFrame. Mandatory columns are determined by the `event_cfg` configuration
            dictionary.
        event_cfg: A dictionary containing the configuration for the event. This must contain the critical
            `"code"` key alongside a mandatory `_metadata` block, which must contain some columns that should
            be extracted from the metadata to link to the code.
            The `"code"` key must contain either (1) a string literal representing the code for the event or
            (2) the name of a column in the raw data from which the code should be extracted. In the latter
            case, the column name should be enclosed in `col()` function call syntax--e.g.,
            `col(my_code_column)`. Note there are no quotes used inside the `col()` function syntax.

    Returns:
        A DataFrame containing the metadata extracted and linked to appropriately constructed code strings for
        the event configuration. The output DataFrame will contain at least two columns: `"code"` and whatever
        metadata column is specified for extraction in the metadata block. The output dataframe will not
        necessarily be unique by code if the input metadata is not unique by code.

    Raises:
        KeyError: If the event configuration dictionary is missing the `"code"` or `"_metadata"` keys or if
            the `"_metadata_"` key is empty or if columns referenced by the event configuration dictionary are
            not found in the raw metadata.

    Examples:
        >>> extract_metadata(pl.DataFrame(), {})
        Traceback (most recent call last):
            ...
        KeyError: "Event configuration dictionary must contain 'code' key. Got: []."
        >>> extract_metadata(pl.DataFrame(), {"code": "test"})
        Traceback (most recent call last):
            ...
        KeyError: "Event configuration dictionary must contain a non-empty '_metadata' key. Got: [code]."
        >>> raw_metadata = pl.DataFrame({
        ...     "code": ["A", "B", "C", "D", "E"],
        ...     "code_modifier": ["1", "2", "3", "4", "5"],
        ...     "name": ["Code A-1", "B-2", "C with 3", "D, but 4", None],
        ...     "priority": [1, 2, 3, 4, 5],
        ... })
        >>> event_cfg = {
        ...     "code": ["FOO", "col(code)", "col(code_modifier)"],
        ...     "_metadata": {"desc": "name"},
        ... }
        >>> extract_metadata(raw_metadata, event_cfg)
        shape: (4, 2)
        ┌───────────┬──────────┐
        │ code      ┆ desc     │
        │ ---       ┆ ---      │
        │ str       ┆ str      │
        ╞═══════════╪══════════╡
        │ FOO//A//1 ┆ Code A-1 │
        │ FOO//B//2 ┆ B-2      │
        │ FOO//C//3 ┆ C with 3 │
        │ FOO//D//4 ┆ D, but 4 │
        └───────────┴──────────┘
        >>> extract_metadata(raw_metadata, event_cfg, allowed_codes=["FOO//A//1", "FOO//C//3"])
        shape: (2, 2)
        ┌───────────┬──────────┐
        │ code      ┆ desc     │
        │ ---       ┆ ---      │
        │ str       ┆ str      │
        ╞═══════════╪══════════╡
        │ FOO//A//1 ┆ Code A-1 │
        │ FOO//C//3 ┆ C with 3 │
        └───────────┴──────────┘
        >>> extract_metadata(raw_metadata.drop("code_modifier"), event_cfg)
        Traceback (most recent call last):
            ...
        KeyError: "Columns {'code_modifier'} not found in metadata columns: ['code', 'name', 'priority']"
        >>> extract_metadata(raw_metadata, ['foo'])
        Traceback (most recent call last):
            ...
        TypeError: Event configuration must be a dictionary. Got: <class 'list'> ['foo'].

    You can also manipulate the columns in more complex ways when assigning metadata from the input source,
    and mandatory MEDS metadata columns will be cast to the correct types:
        >>> raw_metadata = pl.DataFrame({
        ...     "code": ["A", "A", "C", "D"],
        ...     "code_modifier": ["1", "1", "2", "3"],
        ...     "code_modifier_2": ["1", "2", "3", "4"],
        ...     "title": ["A-1-1", "A-1-2", "C-2-3", None],
        ...     "special_title": ["used", None, None, None],
        ... })
        >>> event_cfg = {
        ...     "code": ["FOO", "col(code)", "col(code_modifier)"],
        ...     "_metadata": {
        ...         "description": ["special_title", "title"],
        ...         "parent_codes": [
        ...             {"OUT_VAL/{code_modifier}/2": {"code_modifier_2": "2"}},
        ...             {"OUT_VAL_for_3/{code_modifier}": {"code_modifier_2": "3"}},
        ...             {
        ...                 "matcher": {"code_modifier_2": "4"},
        ...                 "output": {"literal": "expanded form"},
        ...             },
        ...         ],
        ...     },
        ... }
        >>> extract_metadata(raw_metadata, event_cfg)
        shape: (4, 3)
        ┌───────────┬─────────────┬─────────────────────┐
        │ code      ┆ description ┆ parent_codes        │
        │ ---       ┆ ---         ┆ ---                 │
        │ str       ┆ str         ┆ list[str]           │
        ╞═══════════╪═════════════╪═════════════════════╡
        │ FOO//A//1 ┆ used        ┆ null                │
        │ FOO//A//1 ┆ A-1-2       ┆ ["OUT_VAL/1/2"]     │
        │ FOO//C//2 ┆ C-2-3       ┆ ["OUT_VAL_for_3/2"] │
        │ FOO//D//3 ┆ null        ┆ ["expanded form"]   │
        └───────────┴─────────────┴─────────────────────┘
    """
    event_cfg = copy.deepcopy(event_cfg)

    if not isinstance(event_cfg, (dict, DictConfig)):
        raise TypeError(f"Event configuration must be a dictionary. Got: {type(event_cfg)} {event_cfg}.")

    if "code" not in event_cfg:
        raise KeyError(
            "Event configuration dictionary must contain 'code' key. "
            f"Got: [{', '.join(event_cfg.keys())}]."
        )
    if "_metadata" not in event_cfg or not event_cfg["_metadata"]:
        raise KeyError(
            "Event configuration dictionary must contain a non-empty '_metadata' key. "
            f"Got: [{', '.join(event_cfg.keys())}]."
        )

    df_select_exprs = {}
    final_cols = []
    needed_cols = set()
    for out_col, in_cfg in event_cfg["_metadata"].items():
        in_expr, needed = cfg_to_expr(in_cfg)
        df_select_exprs[out_col] = in_expr
        final_cols.append(out_col)
        needed_cols.update(needed)

    code_expr, _, needed_code_cols = get_code_expr(event_cfg.pop("code"))

    columns = metadata_df.collect_schema().names()
    missing_cols = (needed_cols | needed_code_cols) - set(columns) - set(final_cols)
    if missing_cols:
        raise KeyError(f"Columns {missing_cols} not found in metadata columns: {columns}")

    for col in needed_code_cols:
        if col not in df_select_exprs:
            df_select_exprs[col] = pl.col(col)

    metadata_df = metadata_df.select(**df_select_exprs).with_columns(code=code_expr)

    if allowed_codes:
        metadata_df = metadata_df.filter(pl.col("code").is_in(allowed_codes))

    metadata_df = metadata_df.filter(~pl.all_horizontal(*[pl.col(c).is_null() for c in final_cols]))

    for mandatory_col, mandatory_type in MEDS_METADATA_MANDATORY_TYPES.items():
        if mandatory_col not in final_cols:
            continue

        if metadata_df.schema[mandatory_col] is not mandatory_type:
            logger.warning(f"Metadata column '{mandatory_col}' must be of type {mandatory_type}. Casting.")
            metadata_df = metadata_df.with_columns(pl.col(mandatory_col).cast(mandatory_type), strict=False)

    return metadata_df.unique(maintain_order=True).select("code", *final_cols)


def extract_all_metadata(
    metadata_df: pl.LazyFrame, event_cfgs: list[dict], allowed_codes: list | None = None
) -> pl.LazyFrame:
    """Extracts all metadata for a list of event configurations.

    Args:
        metadata_df: The raw metadata DataFrame. Mandatory columns are determined by the `event_cfg`
            configurations.
        event_cfgs: A list of event configuration dictionaries. Each dictionary must contain the code
            and metadata elements.
        allowed_codes: A list of codes to allow in the output metadata. If None, all codes are allowed.

    Returns:
        A unified DF containing all metadata for all event configurations.

    Examples:
        >>> raw_metadata = pl.DataFrame({
        ...     "code": ["A", "B", "C", "D"],
        ...     "code_modifier": ["1", "2", "3", "4"],
        ...     "name": ["Code A-1", "B-2", "C with 3", "D, but 4"],
        ...     "priority": [1, 2, 3, 4],
        ... })
        >>> event_cfg_1 = {
        ...     "code": ["FOO", "col(code)", "col(code_modifier)"],
        ...     "_metadata": {"desc": "name"},
        ... }
        >>> event_cfg_2 = {
        ...     "code": ["BAR", "col(code)", "col(code_modifier)"],
        ...     "_metadata": {"desc2": "name"},
        ... }
        >>> event_cfgs = [event_cfg_1, event_cfg_2]
        >>> extract_all_metadata(raw_metadata, event_cfgs, allowed_codes=["FOO//A//1", "BAR//B//2"])
        shape: (2, 3)
        ┌───────────┬──────────┬───────┐
        │ code      ┆ desc     ┆ desc2 │
        │ ---       ┆ ---      ┆ ---   │
        │ str       ┆ str      ┆ str   │
        ╞═══════════╪══════════╪═══════╡
        │ FOO//A//1 ┆ Code A-1 ┆ null  │
        │ BAR//B//2 ┆ null     ┆ B-2   │
        └───────────┴──────────┴───────┘
    """

    all_metadata = []
    for event_cfg in event_cfgs:
        all_metadata.append(extract_metadata(metadata_df, event_cfg, allowed_codes=allowed_codes))

    return pl.concat(all_metadata, how="diagonal_relaxed").unique(maintain_order=True)


def get_events_and_metadata_by_metadata_fp(event_configs: dict | DictConfig) -> dict[str, dict[str, dict]]:
    """Reformats the event conversion config to map metadata file input prefixes to linked event configs.

    Args:
        event_configs: The event conversion configuration dictionary.

    Returns:
        A dictionary keyed by metadata input file prefix mapping to a dictionary of event configurations that
        link to that metadata prefix.

    Examples:
        >>> event_configs = {
        ...     "subject_id_col": "MRN",
        ...     "icu/procedureevents": {
        ...         "subject_id_col": "subject_id",
        ...         "start": {
        ...             "code": ["PROCEDURE", "START", "col(itemid)"],
        ...             "_metadata": {
        ...                 "proc_datetimeevents": {"desc": ["omop_concept_name", "label"]},
        ...                 "proc_itemid": {"desc": ["omop_concept_name", "label"]},
        ...             },
        ...         },
        ...         "end": {
        ...             "code": ["PROCEDURE", "END", "col(itemid)"],
        ...             "_metadata": {
        ...                 "proc_datetimeevents": {"desc": ["omop_concept_name", "label"]},
        ...                 "proc_itemid": {"desc": ["omop_concept_name", "label"]},
        ...             },
        ...         },
        ...     },
        ...     "icu/inputevents": {
        ...         "event": {
        ...             "code": ["INFUSION", "col(itemid)"],
        ...             "_metadata": {
        ...                 "inputevents_to_rxnorm": {"desc": "{label}", "itemid": "{foo}"}
        ...             },
        ...         },
        ...     },
        ... }
        >>> get_events_and_metadata_by_metadata_fp(event_configs) # doctest: +NORMALIZE_WHITESPACE
        {'proc_datetimeevents': [{'code': ['PROCEDURE', 'START', 'col(itemid)'],
                                  '_metadata': {'desc': ['omop_concept_name', 'label']}},
                                 {'code': ['PROCEDURE', 'END', 'col(itemid)'],
                                  '_metadata': {'desc': ['omop_concept_name', 'label']}}],
         'proc_itemid':         [{'code': ['PROCEDURE', 'START', 'col(itemid)'],
                                  '_metadata': {'desc': ['omop_concept_name', 'label']}},
                                 {'code': ['PROCEDURE', 'END', 'col(itemid)'],
                                  '_metadata': {'desc': ['omop_concept_name', 'label']}}],
         'inputevents_to_rxnorm': [{'code': ['INFUSION', 'col(itemid)'],
                                    '_metadata': {'desc': '{label}', 'itemid': '{foo}'}}]}
        >>> no_metadata_event_configs = {
        ...     "icu/procedureevents": {
        ...         "start": {"code": ["PROCEDURE", "START", "col(itemid)"]},
        ...         "end": {"code": ["PROCEDURE", "END", "col(itemid)"]},
        ...     },
        ...     "icu/inputevents": {
        ...         "event": {"code": ["INFUSION", "col(itemid)"]},
        ...     },
        ... }
        >>> get_events_and_metadata_by_metadata_fp(no_metadata_event_configs)
        {}
    """

    out = {}

    for file_pfx, event_cfgs_for_pfx in event_configs.items():
        if file_pfx == "subject_id_col":
            continue

        for event_key, event_cfg in event_cfgs_for_pfx.items():
            if event_key == "subject_id_col":
                continue

            for metadata_pfx, metadata_cfg in event_cfg.get("_metadata", {}).items():
                if metadata_pfx not in out:
                    out[metadata_pfx] = []
                out[metadata_pfx].append({"code": event_cfg["code"], "_metadata": metadata_cfg})

    return out


@hydra.main(version_base=None, config_path=str(CONFIG_YAML.parent), config_name=CONFIG_YAML.stem)
def main(cfg: DictConfig):
    """Extracts any dataset-specific metadata and adds it to any existing code metadata file.

    This script can extract arbitrary, code-linked metadata columns from input mappings and add them to the
    `metadata/codes.parquet` file. The metadata columns are extracted from the raw metadata files using a
    parsing DSL that is specified in the `event_conversion_config_fp` file. See `parser.py` for more details
    on this DSL.

    Note that there are two sentinel columns in the output metadata that have certain mandates for MEDS
    compliance: The `description` column and the `parent_codes` column. The `description` column must be a
    string, and if there are multiple matches in the extracted metadata for a code, in this script they will
    be concatenated into a single string with the `description_separator` string. The `parent_codes` column
    must be a list of strings, each formatted as an OMOP vocabulary name, followed by a "/", followed by the
    OMOP concept code. This column is used to link codes to their parent codes in the OMOP vocabulary.

    All arguments are specified through the command line into the `cfg` object through Hydra.

    The `cfg.stage_cfg` object is a special key that is imputed by OmegaConf to contain the stage-specific
    configuration arguments based on the global, pipeline-level configuration file. It cannot be overwritten
    directly on the command line, but can be overwritten implicitly by overwriting components of the
    `stage_configs.extract_code_metadata` key.

    Args:
        stage_configs.extract_code_metadata.description_separator: If there are multiple metadata matches for
            a row, this string will be used as a separator to join the matches for the sentinel
            `"description"` column into a single string in the output metadata, per compliance with the MEDS
            schema.
    """

    stage_input_dir, partial_metadata_dir, _ = stage_init(cfg)
    raw_input_dir = Path(cfg.input_dir)

    event_conversion_cfg_fp = Path(cfg.event_conversion_config_fp)
    if not event_conversion_cfg_fp.exists():
        raise FileNotFoundError(f"Event conversion config file not found: {event_conversion_cfg_fp}")

    logger.info(f"Reading event conversion config from {event_conversion_cfg_fp}")
    event_conversion_cfg = OmegaConf.load(event_conversion_cfg_fp)
    logger.info(f"Event conversion config:\n{OmegaConf.to_yaml(event_conversion_cfg)}")

    partial_metadata_dir.mkdir(parents=True, exist_ok=True)
    OmegaConf.save(event_conversion_cfg, partial_metadata_dir / "event_conversion_config.yaml")

    events_and_metadata_by_metadata_fp = get_events_and_metadata_by_metadata_fp(event_conversion_cfg)
    if not events_and_metadata_by_metadata_fp:
        logger.info("No _metadata blocks in the event_conversion_config.yaml found. Exiting...")
        return

    event_metadata_configs = list(events_and_metadata_by_metadata_fp.items())
    random.shuffle(event_metadata_configs)

    # Load all codes
    all_codes = (
        pl.scan_parquet(stage_input_dir / "**/*.parquet")
        .select(pl.col("code").unique())
        .collect()
        .get_column("code")
        .to_list()
    )

    all_out_fps = []
    for input_prefix, event_metadata_cfgs in event_metadata_configs:
        event_metadata_cfgs = copy.deepcopy(event_metadata_cfgs)

        metadata_fp, read_fn = get_supported_fp(raw_input_dir, input_prefix)
        if metadata_fp.suffix != ".parquet":
            read_fn = partial(read_fn, infer_schema=False)
        out_fp = partial_metadata_dir / f"{input_prefix}.parquet"
        logger.info(f"Extracting metadata from {metadata_fp} and saving to {out_fp}")

        compute_fn = partial(extract_all_metadata, event_cfgs=event_metadata_cfgs, allowed_codes=all_codes)

        rwlock_wrap(metadata_fp, out_fp, read_fn, write_lazyframe, compute_fn, do_overwrite=cfg.do_overwrite)
        all_out_fps.append(out_fp)

    logger.info("Extracted metadata for all events. Merging.")

    if cfg.worker != 0:  # pragma: no cover
        logger.info("Code metadata extraction completed. Exiting")
        return

    logger.info("Starting reduction process")

    while not all(fp.exists() for fp in all_out_fps):  # pragma: no cover
        missing_files_str = "\n".join(f"  - {str(fp.resolve())}" for fp in all_out_fps if not fp.exists())
        logger.info("Waiting to begin reduction for all files to be written...\n" f"{missing_files_str}")
        time.sleep(cfg.polling_time)

    start = datetime.now()
    logger.info("All map shards complete! Starting code metadata reduction computation.")

    def reducer_fn(*dfs):
        return pl.concat(dfs, how="diagonal_relaxed").unique(maintain_order=True)

    reduced = reducer_fn(*[pl.scan_parquet(fp, glob=False) for fp in all_out_fps])
    join_cols = ["code", *cfg.get("code_modifier_cols", [])]
    reduced_cols = reduced.collect_schema().names()
    metadata_cols = [c for c in reduced_cols if c not in join_cols]

    n_unique_obs = reduced.select(pl.n_unique(*join_cols)).collect().item()
    n_rows = reduced.select(pl.count()).collect().item()
    logger.info(f"Collected metadata for {n_unique_obs} unique codes among {n_rows} total observations.")

    if n_unique_obs != n_rows:
        aggs = {c: pl.col(c) for c in metadata_cols if c not in MEDS_METADATA_MANDATORY_TYPES}
        if "description" in metadata_cols:
            separator = cfg.stage_cfg.description_separator
            aggs["description"] = pl.col("description").str.concat(separator)
        if "parent_codes" in metadata_cols:
            aggs["parent_codes"] = pl.col("parent_codes").explode()

        reduced = reduced.group_by(join_cols).agg(**aggs)

    reduced = reduced.collect()

    metadata_input_dir = Path(cfg.stage_cfg.metadata_input_dir)
    old_metadata_fp = metadata_input_dir / "codes.parquet"

    if old_metadata_fp.exists():
        logger.info(f"Joining to existing code metadata at {str(old_metadata_fp.resolve())}")
        existing = pl.read_parquet(old_metadata_fp, use_pyarrow=True)
        reduced = existing.join(reduced, on=join_cols, how="full", coalesce=True)

    reducer_fp = Path(cfg.stage_cfg.reducer_output_dir) / "codes.parquet"
    reducer_fp.parent.mkdir(parents=True, exist_ok=True)
    reduced.write_parquet(reducer_fp, use_pyarrow=True)
    logger.info(f"Finished reduction in {datetime.now() - start}")


if __name__ == "__main__":  # pragma: no cover
    main()



import polars as pl

from MEDS_transforms import EXTRACT_CONFIG_YAML, MANDATORY_COLUMNS, MANDATORY_TYPES

# We set this equality explicitly here so linting does not remove an apparently "unused" import if we just
# rename with "as" during the import.
CONFIG_YAML = EXTRACT_CONFIG_YAML

# TODO(mmd): This should really somehow be pulled from MEDS.
MEDS_METADATA_MANDATORY_TYPES = {
    "code": pl.String,
    "description": pl.String,
    "parent_codes": pl.List(pl.String),
}

MEDS_DATA_MANDATORY_TYPES = {c: MANDATORY_TYPES[c] for c in MANDATORY_COLUMNS}



import os

import rootutils

root = rootutils.setup_root(__file__, dotenv=True, pythonpath=True, cwd=True)

code_root = root / "src" / "MEDS_transforms"

USE_LOCAL_SCRIPTS = os.environ.get("DO_USE_LOCAL_SCRIPTS", "0") == "1"

if USE_LOCAL_SCRIPTS:
    RUNNER_SCRIPT = code_root / "runner.py"
else:
    RUNNER_SCRIPT = "MEDS_transform-runner"
